import emfile

from ...utils import _path


def write_em(imageblock, file_path, overwrite=False):
    """
    write an image block to disk as an .em file
    """
    path = str(_path(file_path))
    if not path.endswith('.em'):
        path = f'{path}.em'
    emfile.write(path, imageblock.data, header_params={}, overwrite=overwrite)



from .base import SimpleBlock


class PropertyBlock(SimpleBlock):
    """
    PropertyBlock is a simple dictionary wrapper for arbitrary data
    """
    def _data_setter(self, data):
        return dict(data)

    def items(self):
        return self.data.items()

    @staticmethod
    def _merge_data(datablocks):
        # TODO: this is probably gonna break
        return {k: v for db in datablocks for k, v in db.items()}

    @staticmethod
    def _stack_data(datablocks):
        return {}



from napari.layers import Points
from peepingtom.peeper import PointDepictor

from ...test_data.blocks import pointblock


def test_points_depictor():
    point_depictor = PointDepictor(pointblock, peeper=None)
    assert point_depictor.datablock is pointblock
    point_depictor.init_layers()
    assert len(point_depictor.layers) == 1
    assert isinstance(point_depictor.layers[0], Points)



from pathlib import Path
from typing import Union, List

import mrcfile


def data_from_header(*files: Union[List[str, Path], str, Path], attributes: List[str]):
    """

    Parameters
    ----------
    files
    attributes

    Returns dict of format {filename : attributes} for each mrcfile
    -------

    """
    data = {}
    files = [files] if isinstance(files, (str, Path)) else files

    for file in files:
        with mrcfile.open(file, header_only=True, permissive=True) as mrc:
            _attributes = {attribute: getattr(mrc.header, attribute) for attribute in attributes}
            data[str(file)] = _attributes
    return data


def nx(*files: Union[List[str, Path], str, Path]):
    data = data_from_header(*files, attributes=['nx'])
    return {name: data[name]['nx'] for name in data.keys()}


def ny(*files: Union[List[str, Path], str, Path]):
    data = data_from_header(*files, attributes=['ny'])
    return {name: data[name['ny']] for name in data.keys()}


def nz(*files: Union[List[str, Path], str, Path]):
    data = data_from_header(*files, attributes=['nz'])
    return {name: data[name]['nz'] for name in data.keys()}


def nxnynz(*files: Union[List[str, Path], str, Path]):
    return data_from_header(*files, attributes=['nx', 'ny', 'nz'])



from napari.layers import Shapes, Points
from peepingtom.peeper import LineDepictor

from ...test_data.blocks import lineblock


def test_line_depictor():
    line_depictor = LineDepictor(lineblock, peeper=None)
    assert line_depictor.datablock is lineblock
    line_depictor.init_layers()
    assert len(line_depictor.layers) == 2
    assert isinstance(line_depictor.points_layer, Points)
    assert isinstance(line_depictor.backbone_layer, Shapes)



"""
Gui elements for interfacing peeper and napari
"""

from enum import Enum
from math import floor, ceil
from inspect import signature

import numpy as np
import napari
from magicgui import magicgui
from magicgui._qt.widgets import QDoubleSlider, QDataComboBox
from napari.layers import Layer, Image, Points


class MySlider(QDoubleSlider):
    def setMinimum(self, value: float):
        """Set minimum position of slider in float units."""
        super().setMinimum(int(value * self.PRECISION))


def make_property_widgets(points_layer, property_conditions: dict, output_layer='result'):
    """
    conditions can be strings such as:
        - >, <, >=, <=: continuous properties, make sliders to set the cutoff
        - 'check': property is categorical, make a checkbox for each unique value
    """
    available_conds = ('>', '<', '==', '<=', '>=')
    #sanity check:
    categorical_props = {}
    continuous_props = {}
    for p, c in property_conditions.items():
        if p not in points_layer.properties:
            raise ValueError(f'{p} is not a property of {points_layer}')
        if c == 'check':
            categorical_props[p] = {}
            categorical_props[p]['values'] = np.unique(points_layer.properties[p])
        elif c in available_conds:
            continuous_props[p] = {}
            continuous_props[p]['cond'] = c
            continuous_props[p]['min'] = points_layer.properties[p].min()
            continuous_props[p]['max'] = points_layer.properties[p].max()
        else:
            raise ValueError("conditions can only be 'check' or one of {available_conds}")

    magic_kwargs = {}
    for prop, params in continuous_props.items():
        magic_kwargs[prop] = {'widget_type': MySlider, 'minimum': params['min'], 'maximum': params['max']}
    # programmatically generate function call
    func_lines = []
    func_lines.append(f'def magic_widget(layer: Points, {": float, ".join(continuous_props)}: float) -> Layer:')
    func_lines.append(f'sele = np.ones(len(layer.data))')
    for prop, params in continuous_props.items():
        func_lines.append(f'sele_{prop} = layer.properties["{prop}"] {params["cond"]} {prop}')
        func_lines.append(f'sele = np.logical_and(sele, sele_{prop})')
    func_lines.append(f'return [(layer.data[sele], {{"name": "{output_layer}", "size": 3 }}, "points")]')
    func = '\n    '.join(func_lines)
    # make the widget
    exec(func, globals())
    widget_gui = magicgui(magic_widget, auto_call=True, **magic_kwargs)

    return widget_gui.Gui()


@magicgui(auto_call=True)
def checkboxes(class_0: True, class_1: True, class_2: True):
    pass


class Axis(Enum):
    z = 0
    y = 1
    x = 2

zeros = []

@magicgui(auto_call=True,
          slice_coord={'widget_type': QDoubleSlider, 'fixedWidth': 400, 'maximum': 1},
          mode={'choices': ['average', 'chunk']})
def image_slicer(image: Image, slice_coord: float, slice_size: int, axis: Axis, mode = 'chunk') -> Image:
    im_shape = image.data.shape

    ax_range = im_shape[axis.value] - 1
    ax_range_padded = ax_range - slice_size
    slice_coord_real = int(floor(slice_coord * ax_range_padded + slice_size))
    slice_from = slice_coord_real - floor(slice_size/2)
    slice_to = slice_coord_real + ceil(slice_size/2)

    global zeros
    zeros[:] = 0
    if mode == 'chunk':
        if axis.value == 0:
            zeros[slice_from:slice_to,:,:] = image.data[slice_from:slice_to,:,:]
        elif axis.value == 1:
            zeros[:,slice_from:slice_to,:] = image.data[:,slice_from:slice_to,:]
        elif axis.value == 2:
            zeros[:,:,slice_from:slice_to] = image.data[:,:,slice_from:slice_to]
    elif mode == 'average':
        if axis.value == 0:
            zeros[slice_coord_real,:,:] = image.data[slice_from:slice_to,:,:].mean(axis=0)
        elif axis.value == 1:
            zeros[:,slice_coord_real,:] = image.data[:,slice_from:slice_to,:].mean(axis=1)
        elif axis.value == 2:
            zeros[:,:,slice_coord_real] = image.data[:,:,slice_from:slice_to].mean(axis=2)

    return zeros


def add_widgets(viewer):
    widget = image_slicer.Gui()
    viewer.window.add_dock_widget(widget)
    viewer.layers.events.changed.connect(lambda x: widget.refresh_choices('image'))
    layer_selection_widget = widget.findChild(QDataComboBox, 'image')
    layer_selection_widget.currentTextChanged.connect()



import numpy as np

from peepingtom.core.datablocks import ParticleBlock, PointBlock, OrientationBlock, PropertyBlock
from peepingtom.analysis.particles import classify_radial_distance


pt = PointBlock(np.zeros((10, 3)))
ori = OrientationBlock(np.zeros((10, 3, 3)))
prop = PropertyBlock({})


def test_classify_radial_distance():
    p = ParticleBlock(pt, ori, prop)
    centroids, classes = classify_radial_distance([p], n_classes=1, inplace=False)
    assert isinstance(centroids, np.ndarray)
    assert len(centroids) == 1
    assert isinstance(classes, np.ndarray)
    assert len(classes) == len(pt)


def test_classify_radial_distance_inplace():
    p = ParticleBlock(pt, ori, prop)
    classify_radial_distance([p], n_classes=1)
    added_props = ['class_radial', 'class_radial_params']
    assert all([x in p.properties for x in added_props])



from typing import List
from collections.abc import Iterable

from ...utils import AttributedList


class DataBlock:
    """
    Base class for all simple and complex datablocks.
    Provides common methods and easy type inference
    """
    def __init__(self, name=None, parent=None, depictor=None):
        if name is None:
            name = 'NoName'
        self.name = name
        self.parent = parent
        self.depictor = depictor

    def dump(self):
        kwargs = {}
        kwargs.update({'parent': self.parent, 'depictor': self.depictor})
        return kwargs

    def updated(self):
        """
        this function is called when the data changed in order to trigger callbacks
        """
        if self.depictor is not None:
            self.depictor.update()

    def __newlike__(self, *args, **kwargs):
        # this makes sure that operators get the right output in case
        # _merge or _stack return notimplemented
        if args:
            if args[0] is NotImplemented:
                return NotImplemented
        cls = type(self)
        return cls(parent=self.parent, *args, **kwargs)

    def __shape_repr__(self):
        return ''

    def __base_repr__(self):
        return f'{type(self).__name__}[{self.name}]{self.__shape_repr__()}'

    def __repr__(self):
        return f'<{self.__base_repr__()}>'

    def __and__(self, other):
        if isinstance(other, DataBlock):
            return DataCrate([self, other])
        elif isinstance(other, DataCrate):
            return DataCrate(self) + other
        else:
            return NotImplemented

    def __iand__(self, other):
        return NotImplemented

    @staticmethod
    def _merge_data(datablocks):
        """
        convenience method to merge the data of several datablocks
        of the same type into one, within the same ndimensional space
        used by merge and imerge.
        """
        return NotImplemented

    @staticmethod
    def _stack_data(datablocks):
        """
        convenience method to stack the data of several datablocks into one.
        If dimensionality is the same, add a new dimension; otherwise,
        use the next available dimension for the datablocks with smaller dimensionality
        used by stack and istack.
        """
        return NotImplemented

    def _merge(self, datablocks):
        """
        merge several datablocks and return a `newlike` object
        self is not part of merged objects
        """
        return NotImplemented

    def _stack(self, datablocks):
        """
        stack several datablocks and return a `newlike` object
        self is not part of stacked objects
        """
        return NotImplemented

    def _imerge(self, datablocks):
        """
        like merge, but inplace. Self is part of merged objects.
        """
        return NotImplemented

    def _istack(self, datablocks):
        """
        like stack, but inplace. Self is part of stacked objects.
        """
        return NotImplemented

    def __add__(self, other):
        if isinstance(other, type(self)):
            return self._merge([self, other])
        else:
            return NotImplemented

    def __iadd__(self, other):
        if isinstance(other, type(self)):
            self._imerge([other])
            return self
        else:
            return NotImplemented

    def __or__(self, other):
        if isinstance(other, type(self)):
            return self._stack([self, other])
        else:
            return NotImplemented

    def __ior__(self, other):
        if isinstance(other, type(self)):
            self._istack([other])
            return self
        else:
            return NotImplemented


class SimpleBlock(DataBlock):
    """
    Base class for all simple DataBlock objects, data types which can be visualised by Depictors

    SimpleBlock objects must implement a data setter method as _data_setter which returns
    the appropriately formatted data

    Calling __getitem__ on a SimpleBlock will call __getitem__ on its data property
    """

    def __init__(self, data, **kwargs):
        super().__init__(**kwargs)
        self.data = data

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if isinstance(data, type(self)):
            self._data = data.data
        else:
            self._data = self._data_setter(data)
        self.updated()

    def _data_setter(self, data):
        """
        takes raw data and returns it properly formatted to the SimpleBlock subclass specification.
        """
        raise NotImplementedError('SimpleBlocks must implement this method')

    def dump(self):
        kwargs = super().dump()
        kwargs.update({'data': self.data})
        return kwargs

    def __getitem__(self, key):
        return self.data.__getitem__(key)

    def __setitem__(self, key, value):
        self.data.__setitem__(key, value)
        self.updated()

    def __delitem__(self, key):
        self.data.__delitem__(key)
        self.updated()

    def __contains__(self, item):
        return self.data.__contains__(item)

    def __len__(self):
        return self.data.__len__()

    def __iter__(self):
        return self.data.__iter__()

    def __reversed__(self):
        return self.data.__reversed__()

    def _merge(self, datablocks):
        return self.__newlike__(self._merge_data(datablocks))

    def _stack(self, datablocks):
        return self.__newlike__(self._stack_data(datablocks))

    def _imerge(self, datablocks):
        self.data = self._merge_data([self] + datablocks)

    def _istack(self, datablocks):
        self.data = self._stack_data([self] + datablocks)


class MultiBlock(DataBlock):
    """
    Unites multiple SimpleBlocks into a more complex data object

    Note: classes which inherit from 'MultiBlock' should call Super().__init__
    first in their constructors so that references to blocks are correctly defined
    """
    def __init__(self, **kwargs):
        """
        Parameters
        ----------
        kwargs : keyword arguments which get passed down to DataBlock
        """
        super().__init__(**kwargs)
        self.blocks = []

    def __setattr__(self, name, value):
        """
        Extend the functionality of __setattr__ to automatically add datablocks to the
        'blocks' attribute of a 'MultiBlock' when set
        """
        if isinstance(value, SimpleBlock):
            self._add_block(value)
        super().__setattr__(name, value)

    @property
    def blocks(self):
        return self._blocks

    @blocks.setter
    def blocks(self, blocks: List[SimpleBlock]):
        if not isinstance(blocks, list):
            raise ValueError("blocks in a multiblock object must be a list of 'SimpleBlock' objects")
        self._blocks = blocks

    def _add_block(self, block: SimpleBlock):
        """
        Adds a block to an existing list of SimpleBlocks in a MultiBlock

        This is particularly useful when extending the functionality of an existing
        MultiBlock object by inheritance
        """
        self._blocks.append(block)

    @staticmethod
    def _merge_data(multiblocks):
        blocks_data = []
        blocks_all = [mb.blocks for mb in multiblocks]
        # cryptic loop example: datablock types in "blocks" (a, b, c),
        # this loops through the list [(a1, a2, ...), (b1, b2, ...), (c1, c2, ...)]
        # so this separates the components of a list of multiblocks into a lists of
        # simple datablocks of the same type
        for blocks_by_type in zip(*blocks_all):
            blocks_data.append(blocks_by_type[0]._merge_data(blocks_by_type))
        return blocks_data

    @staticmethod
    def _stack_data(multiblocks):
        blocks_data = []
        blocks_all = [mb.blocks for mb in multiblocks]
        # cryptic loop example: datablock types in "blocks" (a, b, c),
        # this loops through the list [(a1, a2, ...), (b1, b2, ...), (c1, c2, ...)]
        # so this separates the components of a list of multiblocks into a lists of
        # simple datablocks of the same type
        for blocks_by_type in zip(*blocks_all):
            blocks_data.append(blocks_by_type[0]._stack_data(blocks_by_type))
        return blocks_data

    def _merge(self, multiblocks):
        return self.__newlike__(*self._merge_data(multiblocks))

    def _stack(self, multiblocks):
        return self.__newlike__(*self._stack_data(multiblocks))

    def _imerge(self, multiblocks):
        new_data = self._merge_data([self] + multiblocks)
        for block, data in zip(self.blocks, new_data):
            block.data = data

    def _istack(self, multiblocks):
        new_data = self._stack_data([self] + multiblocks)
        for block, data in zip(self.blocks, new_data):
            block.data = data


class DataCrate(AttributedList):
    """
    A container for DataBlock objects which exist within the same n-dimensional reference space
    """
    def __init__(self, iterable_or_datablock=()):
        # recursively unpack the iterable into datablocks only
        def unpack(iterable):
            datablocks = []
            if isinstance(iterable, (list, tuple)):
                for item in iterable:
                    datablocks.extend(unpack(item))
            else:
                datablocks.append(iterable)
            return datablocks

        items = unpack(iterable_or_datablock)
        self._checktypes(items)
        super().__init__(items)

    @staticmethod
    def _checktypes(items):
        for item in items:
            if not isinstance(item, DataBlock):
                raise TypeError(f'DataCrate can only hold DataBlocks, not {type(item)}')

    def __add__(self, other):
        if isinstance(other, list):
            self._checktypes(other)
            return DataCrate(super().__add__(other))
        if isinstance(other, DataBlock):
            return self + DataCrate([other])
        else:
            return NotImplemented

    def __iadd__(self, other):
        if isinstance(other, list):
            self._checktypes(other)
            super().__iadd__(other)
            return self
        if isinstance(other, DataBlock):
            super().__iadd__([other])
            return self
        else:
            return NotImplemented

    def __and__(self, other):
        if isinstance(other, (list, DataBlock)):
            return self + other
        else:
            return NotImplemented

    def __iand__(self, other):
        if isinstance(other, (list, DataBlock)):
            self += other
            return self
        else:
            return NotImplemented

    def __base_repr__(self):
        return f'{type(self).__name__}({len(self)})'

    def __repr__(self):
        return f'<{self.__base_repr__()}: [{", ".join([datablock.__base_repr__() for datablock in self])}]>'



import numpy as np

from .base import SimpleBlock


class ImageBlock(SimpleBlock):
    """
    n-dimensional image block
    data can be interpreted as n-dimensional images
    """

    def __init__(self, data, pixel_size=1, **kwargs):
        super().__init__(data, **kwargs)
        self.pixel_size = pixel_size

    def _data_setter(self, image: np.ndarray):
        return np.array(image)

    @property
    def pixel_size(self):
        return self._pixel_size

    @pixel_size.setter
    def pixel_size(self, value):
        self._pixel_size = float(value)

    def ndim(self):
        return self.data.ndim

    def dump(self):
        kwargs = super().dump()
        kwargs.update({'data': self.data})
        return kwargs

    @staticmethod
    def _stack_data(datablocks):
        return np.stack([db.data for db in datablocks])

    def __shape_repr__(self):
        return f'{self.data.shape}'



from .base import Depictor


class PointDepictor(Depictor):
    def init_layers(self, point_kwargs={}, vector_kwargs={}):
        pkwargs = {'size': 3}
        vkwargs = {'length': 10}

        pkwargs.update(point_kwargs)
        vkwargs.update(vector_kwargs)

        layer = self.make_points_layer(self.datablock.as_zyx(),
                                       name=f'{self.name}',
                                       **pkwargs)
        self.layers.append(layer)

    def push_changes(self):
        self.datablock.data = self.layers[0].data



import numpy as np

from ..datablocks import LineBlock, ParticleBlock, PointBlock, OrientationBlock, PropertyBlock
from .base import Alchemist


class PointToLineAlchemist(Alchemist):
    """
    transform a PointBlock into a LineBlock
    """
    def __init__(self, pointblock):
        super().__init__({'points': pointblock})

    def transform(self):
        points = self.inputs['points'].data
        if 'line' not in self.outputs:
            self.outputs['line'] = LineBlock(points)
        else:
            self.outputs['line'].data = points


class PointToParticleAlchemist(Alchemist):
    """
    transform a PointBlock into a ParticleBlock
    """
    def __init__(self, pointblock):
        super().__init__({'points': pointblock})

    def transform(self):
        points = self.inputs['points'].data
        if 'particles' not in self.outputs:
            pointblock = PointBlock(points)
            oriblock = OrientationBlock(np.zeros((pointblock.n, pointblock.ndim, pointblock.ndim)))
            propblock = PropertyBlock({})
            self.outputs['particles'] = ParticleBlock(pointblock, oriblock, propblock)
        else:
            self.outputs['particles'].positions.data = points



from .base import DataBlock, SimpleBlock, MultiBlock, DataCrate
from .dipoleblock import DipoleBlock
from .imageblock import ImageBlock
from .lineblock import LineBlock
from .meshblock import MeshBlock
from .orientationblock import OrientationBlock
from .particleblock import ParticleBlock
from .pointblock import PointBlock
from .propertyblock import PropertyBlock
from .sphereblock import SphereBlock
from .transformblock import TransformBlock



"""
example DataBlocks of each type
"""
import numpy as np

from peepingtom.core import PointBlock, LineBlock, OrientationBlock, ImageBlock, PropertyBlock, ParticleBlock
from peepingtom.utils import align_vectors

# points
n = 100

x = np.random.normal(0, 30, n)
y = np.random.normal(0, 30, n)
z = np.random.normal(0, 50, n)

xyz = np.column_stack([x, y, z])

pointblock = PointBlock(xyz)

# line
v = np.linspace(0.5, 5 * np.pi, n)
x = np.sin(v)
y = np.cos(v)

xyz = np.column_stack([x, y, v])

lineblock = LineBlock(xyz)

# orientations
n = 100
v = np.linspace(0.5, 2 * np.pi, n)
x = np.sin(v)
y = np.cos(v)
z = np.zeros(n)

xyz = np.column_stack([x, y, z])
unit_z = np.asarray([0, 0, 1])
rotation_matrices = align_vectors(unit_z, xyz)

orientationblock = OrientationBlock(rotation_matrices)

# properties
n = 100
p_dict = {'p1': np.ones((n))}

propertyblock = PropertyBlock(p_dict)

# particles
particleblock = ParticleBlock(pointblock, orientationblock, propertyblock)

# image
img = np.ones((10, 10, 10))
imageblock = ImageBlock(img)



"""
Analysis functions that operate on collections of data object
"""

import numpy as np
from scipy.spatial import cKDTree
from scipy.ndimage import convolve1d
from scipy.cluster.vq import kmeans2
from scipy.signal.windows import gaussian


def classify_radial_distance(particles, class_tag='class_radial', max_r=50, n_shells=100,
                             n_classes=5, convolve=True, cv_window=20, std=5, inplace=True,
                             use_old_matrix=True):
    """
    classify particles based on the number of neighbors in radial shells
    """
    shell_width = max_r / n_shells
    binned = []
    # calculate adjacency matrix or reuse old one
    for part in particles:
        redo = True
        if f'{class_tag}_params' in part.properties:
            old_max_r = part.properties[f'{class_tag}_params']['max_r']
            if max_r <= old_max_r:
                adj_matrix = part.properties[f'{class_tag}_adj_matrix']
                redo = False
        if not use_old_matrix or redo:
            tree = cKDTree(part.positions.data)
            adj_matrix = tree.sparse_distance_matrix(tree, max_r).toarray()
            if inplace:
                part.properties[f'{class_tag}_adj_matrix'] = adj_matrix

        # calculate shells

        shells = [np.sum((adj_matrix > i * shell_width) & (adj_matrix <= (i + 1) * shell_width), axis=1)
                  for i in range(n_shells)]
        binned.append(np.stack(shells, axis=1).astype(float))
    binned = np.concatenate(binned)

    if convolve:
        binned = convolve1d(binned, gaussian(cv_window, std))
    centroids, classes = kmeans2(binned, n_classes, iter=100, minit='points')

    if inplace:
        start = end = 0
        for part in particles:
            end += part.positions.data.shape[0]
            sliced_classes = classes[start:end]
            part.properties[class_tag] = sliced_classes
            start += part.positions.data.shape[0]

            part.properties[f'{class_tag}_centroids'] = centroids
            part.properties[f'{class_tag}_params'] = {
                'max_r': max_r
            }

    return centroids, classes



import numpy as np
import pandas as pd
import starfile

from peepingtom.utils.helpers.eulerangle_helper import EulerAngleHelper
from peepingtom.utils.helpers.linalg_helper import align_vectors

from .constants import test_data_dir

# relion star file simple
# coords for helix along z
z = np.linspace(0.1, 6 * np.pi + 0.1, 50)
x = 3 * np.sin(z)
y = 3 * np.cos(z)

# backbone along center of helix
x_backbone = np.zeros(z.shape[0])
y_backbone = np.zeros(z.shape[0])

xyz = np.column_stack([x, y, z])
xyz_backbone = np.column_stack([x_backbone, y_backbone, z])

# calculate directions from backbone to points on helix
orientation_vectors = xyz - xyz_backbone
orientation_vectors_norm = orientation_vectors / np.linalg.norm(orientation_vectors, axis=1).reshape((-1, 1))
unit_z = np.array([0, 0, 1])

# calculate rotation matrices to align unit vector to direction away from backbone
rotation_matrices = align_vectors(unit_z, orientation_vectors_norm)

# calculate eulers from rotation matrices
euler_angles_rln = EulerAngleHelper(rotation_matrices=rotation_matrices).matrix2euler('relion')

rot = euler_angles_rln[:, 0]
tilt = euler_angles_rln[:, 1]
psi = euler_angles_rln[:, 2]

star_info = {'rlnCoordinateX': x,
             'rlnCoordinateY': y,
             'rlnCoordinateZ': z,
             'rlnAngleRot': rot,
             'rlnAngleTilt': tilt,
             'rlnAnglePsi': psi}

star = pd.DataFrame.from_dict(star_info)
star['rlnMicrographName'] = 'test_tomo'
star_path = test_data_dir / 'relion_3d_simple.star'
starfile.write(star, star_path)



def merge(datablocks):
    """
    merges a list of datablocks into a single datablock
    """
    cls = type(datablocks[0])
    if not all(isinstance(db, cls) for db in datablocks):
        raise TypeError('cannot merge datablocks of different types')
    return datablocks[0]._merge(datablocks)


def stack(datablocks):
    """
    stacks a list of datablocks into a single datablock
    """
    cls = type(datablocks[0])
    if not all(isinstance(db, cls) for db in datablocks):
        raise TypeError('cannot stack datablocks of different types')
    return datablocks[0]._stack(datablocks)



import numpy as np
from peepingtom.core.datablocks import ImageBlock


def test_imageblock_instantiation():
    im = np.random.normal(0, 1, (28, 28, 28))
    block = ImageBlock(im)
    assert isinstance(block, ImageBlock)



import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal

from peepingtom.core import DipoleBlock

test_centers = np.asarray([[0, 0, 0],
                           [1, 1, 1],
                           [2, 2, 2]])

test_vectors = np.asarray([[0, 0, 1], [0, 0, 2], [3, 3, 3]])

test_vector_norms = np.expand_dims(np.linalg.norm(test_vectors, axis=1), axis=-1)
normalised_test_vectors = test_vectors / test_vector_norms

test_endpoints = test_centers + test_vectors


def test_dipoleblock_instantiation():
    block = DipoleBlock(test_centers, test_endpoints)
    assert isinstance(block, DipoleBlock)


def test_dipoleblock_orientation_vectors():
    block = DipoleBlock(test_centers, test_endpoints)
    assert_array_equal(block.orientation_vectors, test_vectors)


def test_dipoleblock_normalised_orientation_vector():
    block = DipoleBlock(test_centers, test_endpoints)
    assert_array_almost_equal(block.normalised_orientation_vectors, normalised_test_vectors)


def test_dipoleblock_rotation_matrices():
    block = DipoleBlock(test_centers, test_endpoints)

    # calculate rotation matrices to align test vector to each dipole in block
    vector_to_align = np.array([0, 2, 1]).reshape(1, 3, 1)
    rotation_matrices = block.calculate_orientation_block(vector_to_align)

    # calculate vector to aligns new position after rotation by rotation_matrix
    aligned_vectors = rotation_matrices @ vector_to_align
    aligned_vectors = aligned_vectors.reshape(-1, 3)

    # calculate normalised aligned vectors for comparison with dipoleblock vectors
    aligned_vectors_norm = np.linalg.norm(aligned_vectors, axis=1)
    normalised_aligned_vectors = aligned_vectors / np.expand_dims(aligned_vectors_norm, axis=-1)

    # check normalised aligned vectors match normalised dipoleblock vectors
    assert_array_almost_equal(normalised_aligned_vectors, block.normalised_orientation_vectors)



import pytest
import numpy as np
from numpy.testing import assert_array_equal

from peepingtom.core.datablocks import PointBlock

# test data pointblock
single_point_2d = [1, 2]
points_2d = [[1, 2], [3, 4]]

single_point_3d = [1, 2, 3]
points_3d = [[1, 2, 3], [4, 5, 6]]

point_nd = np.arange(48).reshape(6, 8)


def test_pointblock_instantiation():
    # test instantiation for 2d, 3d and nd points
    PointBlock(single_point_2d)
    PointBlock(points_2d)
    PointBlock(single_point_3d)
    PointBlock(points_3d)
    PointBlock(point_nd)


def test_pointblock_xyz():
    # test 'x', 'y' and 'z' properties
    block = PointBlock(single_point_2d)
    assert_array_equal(block.x, [1])
    assert_array_equal(block.y, [2])

    block = PointBlock(single_point_3d)
    assert_array_equal(block.x, [1])
    assert_array_equal(block.y, [2])
    assert_array_equal(block.z, [3])


def test_pointblock_get_named_dimension():
    # test _get_named_dimension method
    block = PointBlock(single_point_3d)

    # check output is array
    x = block._get_named_dimension('x')
    assert isinstance(x, np.ndarray)

    # try with multiple named dims
    xyz = block._get_named_dimension('xyz')
    assert isinstance(xyz, np.ndarray)
    assert xyz.shape == (1, 3)

    xyz = block._get_named_dimension('xyz', as_type='tuple')
    assert isinstance(xyz, tuple)
    assert len(xyz) == 3


def test_pointblock_center_of_mass():
    # test center_of_mass property
    # single point case
    block = PointBlock(single_point_3d)
    assert_array_equal(block.center_of_mass, single_point_3d)

    # multi point case
    block = PointBlock(points_3d)
    assert_array_equal(block.center_of_mass, [2.5, 3.5, 4.5])


def test_pointblock_distance_to():
    # test distance_to method
    block = PointBlock(single_point_3d)

    assert block.distance_to([1, 2, 3]) == 0
    assert block.distance_to([2, 3, 4]) == np.sqrt(3)

    # check for failure
    with pytest.raises(ValueError):
        block.distance_to([1, 2, 3, 4, 5, 6])



import re
from pathlib import Path
from collections.abc import Iterable


def _path(path):
    """
    clean up a path
    """
    return Path(path).expanduser().resolve()


# a list of commonly used base names for starfiles in regex form
common_name_regexes = (
    'TS_\d+',
)


def guess_name(string, name_regex=None):
    """
    guess an appropriate name based on the input string
    and a list of regexes in order of priority
    """
    regexes = list(common_name_regexes)
    if name_regex is not None:
        regexes.insert(0, name_regex)
    for regex in regexes:
        if match := re.search(regex, str(string)):
            return match.group(0)
    else:
        return None


def listify(obj):
    """
    transform input into an appropriate list, unless already list-like
    """
    if isinstance(obj, (list, tuple)):
        return obj
    if obj is None:
        return []
    return [obj]



from napari.layers import Vectors, Points
from peepingtom.peeper import ParticleDepictor

from ...test_data.blocks import particleblock


def test_particle_depictor():
    particle_depictor = ParticleDepictor(particleblock, peeper=None)
    assert particle_depictor.datablock is particleblock
    particle_depictor.init_layers()
    assert len(particle_depictor.layers) == 2
    assert isinstance(particle_depictor.point_layer, Points)
    assert isinstance(particle_depictor.vector_layer, Vectors)



import emfile

from ....core import ImageBlock
from ...utils import guess_name


def read_em(image_path, name_regex=None, **kwargs):
    """
    read an em file and return an ImageBlock
    """
    name = guess_name(image_path, name_regex)
    return ImageBlock(emfile.read(image_path)[1], name=name)



"""
Depictor interfaces data classes to napari
"""

from napari.layers import Points, Image, Vectors, Shapes

from ...core import MultiBlock


class Depictor:
    """
    Depictors are SimpleBlock or MultiBlock wrappers controlling depiction of their contents
    """
    def __init__(self, datablock, peeper=None):
        self.datablock = datablock

        # hook self to the datablock
        self.datablock.depictor = self
        if isinstance(self.datablock, MultiBlock):
            for block in self.datablock.blocks:
                block.depictor = self

        self.name = self.datablock.name
        self.peeper = peeper

        self.layers = []

    @property
    def viewer(self):
        return self.peeper.viewer

    def make_image_layer(self, image, name, **kwargs):
        return Image(image, name=name, **kwargs)

    def make_points_layer(self, points, name, **kwargs):
        return  Points(points, name=name, **kwargs)

    def make_vectors_layer(self, vectors, name, **kwargs):
        return Vectors(vectors, name=name, **kwargs)

    def make_shapes_layer(self, shape, shape_type, name, **kwargs):
        return Shapes(shape, shape_type=shape_type, name=name, **kwargs)

    def init_layers(self):
        """
        generate the appropriate napari layers and store them
        """
        raise NotImplementedError

    def connect_layers(self):
        """
        connects events on layer data to an update of the relative datablock
        """
        for layer in self.layers:
            layer.events.data.connect(self.push_changes)

    def draw(self, viewer=None, remake_layers=False):
        """
        creates a new napari viewer if not present
        displays the contents of the datablock
        """
        # create a new viewer if necessary
        if viewer is None:
            viewer = self.viewer
        if remake_layers or not self.layers:
            self.init_layers()
            self.connect_layers()
        for layer in self.layers:
            self.viewer.add_layer(layer)

    def hide(self, layers=None, delete_layers=False):
        """
        layer_id can be the index or name of a layer or a list of ids
        """
        if layers is None:
            layers = [layer for layer in self.layers]
        if not isinstance(layers, list):
            layers = [layers]
        for layer in layers:
            if layer in self.layers:
                self.viewer.layers.remove(layer)
                if delete_layers:
                    self.layers.remove(layer)

    def update(self):
        """
        update the displayed data based on the current state of the datablock
        """

    def push_changes(self, event):
        """
        push changes in the local layers to the corresponding datablock
        """

    def __repr__(self):
        return f'<{type(self).__name__}[{self.name}]{self.datablock.__shape_repr__()}>'



from abc import ABC, abstractmethod


class Model(ABC):
    """
    A Model is an object from which a set of Particles can be derived
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def derive_particles(self):
        pass



from .alchemists import Alchemist, PointToLineAlchemist, PointToParticleAlchemist
from .datablocks import DataBlock, SimpleBlock, MultiBlock, DataCrate, PointBlock, \
    LineBlock, OrientationBlock, ImageBlock, LineBlock, SphereBlock, PropertyBlock, \
    ParticleBlock, MeshBlock, DipoleBlock, TransformBlock
from .models import Model

from .functions import merge, stack



import numpy as np

from peepingtom.core import TransformBlock, DipoleBlock, PointBlock, ParticleBlock
from tests.test_data.blocks import particleblock

# test transform
shifts = [
    [-1, 0, 0],
    [-1, 0, 1],
    [-1, 0, -1],
    [-1, 1, 0],
    [-1, -1, 0],
    [0, 0, 0],
    [-2, 0, 0]
]
test_shifts = PointBlock(np.asarray(shifts))

test_endpoints = PointBlock(np.zeros((test_shifts.data.shape[0], 3)))
test_orientations = DipoleBlock(test_shifts, test_endpoints).calculate_orientation_block([0, 0, 1])


def test_transformblock_instantiation():
    # test that block instantiates properly
    block = TransformBlock(test_shifts, test_orientations, {})
    assert isinstance(block, TransformBlock)


def test_transformblock_application():
    block = TransformBlock(test_shifts, test_orientations, {})

    # apply transformations in block on particles
    new_block = block.apply_on(particleblock)

    # check that object returned is ParticleBlock
    assert isinstance(new_block, ParticleBlock)

    # check that no. of new particles = number of transformations * number of particles
    n_transforms = block.positions.data.shape[0]
    n_particles = particleblock.positions.data.shape[0]
    n_transformed_particles = new_block.positions.data.shape[0]

    assert n_transformed_particles == n_transforms * n_particles



import numpy as np
from eulerangles import euler2matrix

from .base import SimpleBlock


class OrientationBlock(SimpleBlock):
    """
    OrientationBlock objects represent orientations in a 2d or 3d space

    Contains factory methods for instantiation from eulerian angles

    rotation_matrices : (n, 2, 2) or (n, 3, 3) array of rotation matrices R
                        R should satisfy Rv = v' where v is a column vector

    """
    def _data_setter(self, data: np.ndarray):
        data = np.array(data)
        # check for single matrix case and assert dimensionality
        val_error = ValueError(f'rotation matrices should be of shape '
                               f'(2, 2), (3, 3), (n, 2, 2) or (n, 3, 3), '
                               f'not {data.shape}')
        if data.ndim == 1:
            raise val_error
        if not data.shape[-1] == data.shape[-2]:
            raise val_error
        if not data.shape[-1] in (2, 3):
            raise val_error

        if data.ndim == 2:
            m = data.shape[-1]
            data = data.reshape((1, m, m))

        return data

    @property
    def n(self):
        return len(self)

    @property
    def ndim(self):
        """
        as ndim for numpy arrays, but treating the vectors as a sparse matrix.
        returns the number of dimensions (spatial or not) describing the vectors
        """
        return self.data.shape[-1]

    @classmethod
    def from_euler_angles(cls, euler_angles: np.ndarray, axes: str, intrinsic: bool, positive_ccw: bool,
                          invert_matrix: bool):
        """
        Factory method for creating a VectorBlock directly from a set of eulerian angles

        Parameters
        ----------
        euler_angles : ndarray, (n, 3) array of n sets of eulerian angles in degrees
        axes : str, 3 characters from 'x', 'y' and 'z' representing axes around which rotations
               occur in the euler angles
        intrinsic : bool, are the euler angles describing a set of intrinsic (rotating reference frame)
                    or extrinsic (fixed reference frame) rotations
        positive_ccw : bool, are positive euler angles referring to counterclockwise rotations of vectors
                       when looking from a positive point along the axis towards the origin
        invert_matrix : bool, should the matrix be inverted?
                        this is useful if your euler angles describe the rotation of a target to a source
                        but you would like your rotation matrices to describe the rotation of a source to
                        align it with a target

        Returns
        -------

        """
        # Calculate rotation matrices
        rotation_matrices = euler2matrix(euler_angles, axes=axes, intrinsic=intrinsic,
                                         positive_ccw=positive_ccw)

        # invert matrix if required
        if invert_matrix:
            rotation_matrices = rotation_matrices.transpose((-1, -2))

        return cls(rotation_matrices)

    def _calculate_matrix_product(self, vector: np.ndarray):
        """
        Calculates the matrix product (v') of the orientation matrices (R) in this
        VectorBlock object with a given vector (v)
        Rv = v'

        Parameters
        ----------
        vector : ndarray v, column vector or set of column vectors to be premultiplied by rotation matrices

        Returns ndarray v', matrix product Rv
        -------

        """
        return self.data @ vector

    def _unit_vector(self, axis: str):
        """
        Get a unit vector along a specified axis which matches the dimensionality of the VectorBlock object

        Parameters
        ----------
        axis : str, named axis 'x', 'y' or 'z'

        Returns unit vector along provided axis with appropriate dimensionality
        -------

        """
        # check dimensionality
        if self.ndim > 3:
            raise NotImplementedError('Unit vector generation for objects with greater '
                                      'than 3 spatial dimensions is not implemented')

        # initialise unit vector array
        unit_vector = np.zeros((self.ndim, 1))

        # get index which corresponds to axis for vector
        axis_to_index = {'x': 0,
                         'y': 1,
                         'z': 2}
        dim_idx = axis_to_index[axis]

        # construct unit vector
        if dim_idx <= self.ndim:
            unit_vector[dim_idx, 0] = 1
        else:
            raise ValueError(f"You asked for axis {axis} from a {self.ndim}d object")

        return unit_vector

    def oriented_vectors(self, axis):
        return self._calculate_matrix_product(self._unit_vector(axis))

    def dump(self):
        kwargs = super().dump()
        kwargs.update({'orientation_block': self.data})
        return kwargs

    @staticmethod
    def _merge_data(datablocks):
        return np.concatenate([db.data for db in datablocks])

    @staticmethod
    def _stack_data(datablocks):
        return np.concatenate([db.data for db in datablocks])

    def __shape_repr__(self):
        return f'({self.n}, {self.ndim})'



from .base import Depictor


class ImageDepictor(Depictor):
    def init_layers(self, image_kwargs={}, **kwargs):
        ikwargs = {}
        ikwargs.update(image_kwargs)

        layer = self.make_image_layer(self.datablock.data,
                                      name=f'{self.name} - image',
                                      **ikwargs)
        self.layers.append(layer)



star_types = {
    'dynamo': {
        'coords': ['x', 'y', 'z'],
        'shifts': ['dx', 'dy', 'dz'],
        'angles': ['tdrot', 'tilt', 'narot'],
        'angle_convention': {'axes': 'zxz',
                             'intrinsic': False,
                             'positive_ccw': True,
                             'rotate_reference': False},
        'split_by': 'rlnMicrographName',
    },
    'relion': {
        'coords': [f'rlnCoordinate{axis}' for axis in 'XYZ'],
        'shifts': [f'rlnOrigin{axis}' for axis in 'XYZ'],
        'angles': [f'rlnAngle{angle}' for angle in ('Rot', 'Tilt', 'Psi')],
        'angle_convention': {'axes': 'zyz',
                             'intrinsic': True,
                             'positive_ccw': True,
                             'rotate_reference': False},
        'split_by': 'rlnMicrographName',
    },
    'relion_2d': {
        'coords': [f'rlnCoordinate{axis}' for axis in 'XY'],
        'shifts': [f'rlnOrigin{axis}' for axis in 'XY'],
    },
}



import numpy as np

from .base import SimpleBlock


class PointBlock(SimpleBlock):
    """
    PointBlock objects for representing points with convenience methods

    PointBlock data should be array-like objects of shape (n, m) representing n points in m dimensions

    order of dimensions along m is:
    2d : (x, y)
    3d : (x. y, z)
    nd : (..., x, y, z)
    """
    def _data_setter(self, data):
        # cast as array
        data = np.asarray(data)

        # coerce 1d point to 2d
        if data.ndim == 1:
            data = data.reshape((1, len(data)))

        # check ndim of data
        if not data.ndim == 2:
            raise ValueError("data object should have ndim == 2")

        return data

    @property
    def ndim(self):
        """
        as ndim for numpy arrays, but treating the points as a sparse matrix.
        returns the number of dimensions (spatial or not) describing the points
        """
        return self.data.shape[1]

    def _named_dimension_to_spatial_index(self, dim: str):
        """
        Gets the index of the named dimension 'x', 'y' or 'z'
        Parameters
        ----------
        dim : str, must be one of 'x', 'y' or 'z'

        Returns data along named dimension
        -------

        """
        # sanitise input
        dim = str(dim.strip().lower())

        # dim to index for 3d or less
        dim_to_index = {'x': 0,
                        'y': 1,
                        'z': 2}

        dim_idx = dim_to_index[dim]

        # check and correct index for higher dimensionality
        if self.ndim > 3:
            dim_idx = -dim_idx - 1

        return dim_idx

    def _get_dim_at_spatial_index(self, idx: int):
        return self[:, idx]

    def _get_named_dimension(self, dim: str, as_type='array'):
        """
        Get data for a named dimension or multiple named dimensions of the object

        as_array and as_tuple are only considered when retrieving multiple dimensions in one method call
        Parameters
        ----------

        dim : str 'x', 'y', 'z' or a combination thereof
        as_type : str for return type, only if len(dim) > 1
                  'array' for ndarray or 'tuple' for tuple return type

        Returns (default) (n,m) ndarray of data along named dimension(s) from m
                  or tuple of arrays of data along each axis
        -------

        """
        if as_type not in ('array', 'tuple'):
            raise ValueError("Argument 'as_type' must be a string from 'array' or 'tuple'")

        if len(dim) > 1:
            # split dims up and get each separately
            data = [self._get_named_dimension(_dim) for _dim in dim]

            # decide on output type and return array or tuple as requested, default to array
            if as_type == 'array':
                return np.column_stack(data)
            elif as_type == 'tuple':
                return tuple(data)

        else:
            # get index of named dimension along spatial axis
            dim_idx = self._named_dimension_to_spatial_index(dim)
            # index into self to get data
            return self._get_dim_at_spatial_index(dim_idx)

    @property
    def x(self):
        return self._get_named_dimension('x')

    @property
    def y(self):
        return self._get_named_dimension('y')

    @property
    def z(self):
        return self._get_named_dimension('z')

    @property
    def xyz(self):
        return self._get_named_dimension('xyz')

    @property
    def zyx(self):
        return self._get_named_dimension('zyx')

    def as_zyx(self):
        """
        return the data with the order of the spatial axes switched to 'zyx' style rather than 'xyz'

        Returns
        -------
        correct view into data no matter the dimensionality
        """
        if self.ndim == 1:
            return self.data
        elif self.ndim == 2:
            # invert last two axes
            return self.data[:, ::-1]
        else:
            # invert only last three axes, leave leading dimensions intact
            data = np.empty_like(self.data)
            data[:, :-3] = self.data[:, :-3]
            data[:, -3:] = self.data[:, -1:-4:-1]
            return data

    @property
    def n(self):
        return len(self)

    @property
    def center_of_mass(self):
        return np.mean(self.data, axis=0)

    def distance_to(self, point):
        """
        Calculate the euclidean distance between the center of mass of this object and a point

        Parameters
        ----------
        point : array-like object

        Returns : euclidean distance
        -------

        """
        point = np.asarray(point)
        if not point.shape == self.center_of_mass.shape:
            raise ValueError(
                f"shape of point '{point.shape}' does not match shape of center of mass "
                f"'{self.center_of_mass.shape}'")
        return np.linalg.norm(point - self.center_of_mass)

    def dump(self):
        kwargs = super().dump()
        kwargs.update({'points': self.data})
        return kwargs

    @staticmethod
    def _merge_data(datablocks):
        return np.concatenate([db.data for db in datablocks])

    @staticmethod
    def _stack_data(datablocks):
        # this method assumes that stacked points are always intended to have
        # indexes at any dimension that's higher than 3
        dims = [db.ndim for db in datablocks]
        # only obvious how to stack elements with just 1 dimension of difference
        if abs(min(dims) - max(dims)) > 1:
            return NotImplemented
        # allocate the right size
        stack_shape = (sum([db.n for db in datablocks]), min(dims) + 1)
        stacked = np.zeros(stack_shape)
        previous = 0
        for idx, db in enumerate(datablocks):
            current = previous + db.n
            if db.ndim == max(dims) and db.ndim != min(dims):
                if len(np.unique(db.data)) != 1:
                    raise ValueError(f'cannot stack: dimension #{db.ndim} of {db} is not an index')
                # if the given data is not in order, just give up. TODO: implement something smarter!
                elif db.data[0, 0] != idx:
                    raise ValueError(f'cannot stack: the index ({db.data[0,0]}) of {db} does not match')
                else:
                    stacked[previous:current] = db.data
            elif db.ndim == min(dims):
                stacked[previous:current] = np.concatenate([np.ones((db.n, 1)) * idx, db.data], axis=1)
            previous = current
        return stacked

    def __shape_repr__(self):
        return f'{self.data.shape}'



"""
main class that interfaces visualization, analysis and data manipulation
"""

import napari

from ...core import ImageBlock, ParticleBlock, PointBlock, LineBlock
from ..depictors import ImageDepictor, ParticleDepictor, PointDepictor, LineDepictor
from ...utils import AttributedList
from ...io_ import read, write


class Peeper:
    """
    collect and display an arbitrary set of images and/or datasets
    expose the datasets to visualization and analysis tools
    """
    def __init__(self, crates, viewer=None):
        self.crates = AttributedList(crates)
        self.viewer = viewer
        # initialise depictors
        for crate in crates:
            for datablock in crate:
                self._init_depictor(datablock)

    def _init_depictor(self, datablock):
        depictor_type = {
            ParticleBlock: ParticleDepictor,
            ImageBlock: ImageDepictor,
            PointBlock: PointDepictor,
            LineBlock: LineDepictor,
        }
        try:
            # don't store a reference to it, cause it hooks itself on the datablock
            depictor_type[type(datablock)](datablock, peeper=self)
        except KeyError:
            raise TypeError(f'cannot find a Depictor for datablock of type {type(datablock)}')

    @property
    def datablocks(self):
        return AttributedList(datablock for crate in self.crates for datablock in crate)

    @property
    def depictors(self):
        return self.datablocks.depictor

    @property
    def depictor_layers(self):
        return self.depictors.layers

    def _get_datablocks(self, block_type):
        return AttributedList(datablock for datablock in self.datablocks if isinstance(datablock, block_type))

    def _init_viewer(self, viewer=None):
        # create a new viewer if necessary
        if viewer is not None:
            self.viewer = viewer
        elif self.viewer is None:
            self.viewer = napari.Viewer(ndisplay=3)

        # random check to make sure viewer was not closed
        try:
            self.viewer.window.qt_viewer.actions()
        except RuntimeError:
            self.viewer = napari.Viewer(ndisplay=3)

    def peep(self, viewer=None):
        self._init_viewer(viewer)
        self.depictors.draw()

    def hide(self):
        self.depictors.hide()

    def update(self):
        self.depictors.update()

    def read(self, paths, **kwargs):
        """
        read paths into datablocks and append them to the datacrates
        """
        self.crates.append(read(paths, **kwargs))

    def write(self, paths, **kwargs):
        """
        write datablock contents to disk
        """
        write(self.datablocks, paths, **kwargs)


def peep(paths, force_mode=None, **kwargs):
    """
    load path(s) as DataCrates into a Peeper object and display them in napari
    """
    peeper = Peeper(read(paths, mode=force_mode, **kwargs))
    peeper.peep()
    return peeper



import numpy as np
from numpy.testing import assert_array_equal
from eulerangles import euler2matrix

from peepingtom.core.datablocks import OrientationBlock

# test data for orientationblock
# orientations
rot = np.zeros(50)
tilt = np.linspace(0, 180, 50)
psi = rot
eulers = np.column_stack([rot, tilt, psi])
rotation_matrices = euler2matrix(eulers, axes='zyz', intrinsic=True, positive_ccw=True)


def test_orientationblock_instantiation():
    block = OrientationBlock(rotation_matrices)
    assert isinstance(block, OrientationBlock)


def test_orientationblock_unit_vectors():
    block = OrientationBlock(rotation_matrices)
    x = block._unit_vector('x')
    y = block._unit_vector('y')
    z = block._unit_vector('z')

    unit_x = np.asarray([1, 0, 0]).reshape(3, 1)
    unit_y = np.asarray([0, 1, 0]).reshape(3, 1)
    unit_z = np.asarray([0, 0, 1]).reshape(3, 1)

    assert_array_equal(x, unit_x)
    assert_array_equal(y, unit_y)
    assert_array_equal(z, unit_z)



import numpy as np

from .base import MultiBlock
from .pointblock import PointBlock
from .orientationblock import OrientationBlock
from ...utils import align_vectors


class DipoleBlock(MultiBlock):
    """
    A DipoleBlock represents a set of n dipoles in an m-dimensional space
    defined by start points and end points
    """

    def __init__(self, startpoints: np.ndarray, endpoints: np.ndarray, **kwargs):
        super().__init__(**kwargs)
        # Set dipole info
        self.startpoints = PointBlock(startpoints)
        self.endpoints = PointBlock(endpoints)

        # Check endpoint shape matches that of startpoints
        self._check_shapes()

    @property
    def orientation_vectors(self):
        """
        Vectors describing shifts to apply on start points to arrive at end points
        """
        return self.endpoints.data - self.startpoints.data

    @property
    def normalised_orientation_vectors(self):
        """
        self.orientation_vectors normalised to length 1
        """
        # Calculate length of each vector, add trailing dimension
        norm = np.expand_dims(np.linalg.norm(self.orientation_vectors, axis=1), axis=-1)
        # Calculate and return normalised vectors
        return self.orientation_vectors / norm

    def _check_shapes(self):
        if self.startpoints.data.shape != self.endpoints.data.shape:
            raise ValueError('Shapes of centers and endpoints do not match')
        else:
            return True

    def calculate_orientation_block(self, vector: np.ndarray):
        """
        Parameters
        -------
        vector : np.ndarray
                 A vector which you would like to align with the orientation vectors of each dipole

        Returns
        -------
        orientation_block : OrientationBlock
                            SimpleBlock containint rotation matrices which premultiply 'vector' to
                            align it with the orientation vectors of each dipole in this object
        """
        # only implemented for 3d rotations
        if self.startpoints.data.shape[1] != 3:
            raise NotImplementedError

        # force ndarray
        vector = np.asarray(vector).reshape(-1)

        # normalise vector
        normalised_vector = vector / np.linalg.norm(vector)

        # calculate rotation matrices
        rotation_matrices = align_vectors(normalised_vector, self.normalised_orientation_vectors)
        return OrientationBlock(rotation_matrices, parent=self)



from napari.layers import Image
from peepingtom.peeper import ImageDepictor

from ...test_data.blocks import imageblock


def test_image_depictor():
    image_depictor = ImageDepictor(imageblock, peeper=None)
    assert image_depictor.datablock is imageblock
    image_depictor.init_layers()
    assert len(image_depictor.layers) == 1
    assert isinstance(image_depictor.layers[0], Image)



import numpy as np
from scipy.interpolate import splprep, splev

from .pointblock import PointBlock


class LineBlock(PointBlock):
    """
    LineBlock objects represent lines with convenience methods

    LineBlock line data should be array-like objects of shape (n, m) representing
    n ordered points in m spatial dimensions

    order of dimensions along m is:
    2d : (x, y)
    3d : (x. y, z)
    nd : (..., x, y, z)

    Polarity (direction) of lines, lines start from 0 to n along the 0th dimension
    """
    def __init__(self, data, spline_smoothing_parameter=0, **kwargs):
        """

        Parameters
        ----------
        data : array-like objects of shape (n, m) representing n ordered points
        in m spatial dimensions
        kwargs : kwargs are passed to SimpleBlock object

        """
        super().__init__(data, **kwargs)

        # initialise attributes related to spline fitting
        self.spline_smoothing_parameter = spline_smoothing_parameter
        self._tck = None

    @property
    def spline_smoothing_parameter(self):
        return self._spline_smoothing_parameter

    @spline_smoothing_parameter.setter
    def spline_smoothing_parameter(self, value):
        self._spline_smoothing_parameter = float(value)

    def fit_spline(self, dimensions: str = 'xyz', smoothing_parameter=None):
        """

        Parameters
        ----------
        dimensions :  str of named dimensions ('xyz') to which a spline should be fit
        smoothing_parameter : smoothing parameter for spline fitting

        Returns tck, list of spline parameters from scipy.interpolate.splprep
        -------

        """
        if smoothing_parameter:
            self.spline_smoothing_parameter = smoothing_parameter

        dims_to_fit = self._get_named_dimension(dimensions, as_type='tuple')
        self._tck, _ = splprep(dims_to_fit, s=self.spline_smoothing_parameter)

        return self._tck

    def evaluate_spline(self, n_points):
        u = np.linspace(0, 1, n_points, endpoint=True)
        return np.asarray(splev(u, tck=self._tck)).T

    @property
    def smooth_backbone(self):
        return self._generate_smooth_backbone()

    def _generate_smooth_backbone(self, n_points=1000):
        self.fit_spline()
        return self.evaluate_spline(n_points)

    def dump(self):
        kwargs = super().dump()
        kwargs.update({'line': self.data})
        return kwargs



import mrcfile

from ....core import ImageBlock
from ...utils import guess_name


def read_mrc(image_path, name_regex=None, **kwargs):
    """
    read an mrc file and return an ImageBlock
    """
    name = guess_name(image_path, name_regex)
    return ImageBlock(mrcfile.open(image_path).data, name=name)



import numpy as np

from peepingtom.core import MeshBlock

vertices = np.array([[0, 0, 0],
                     [0, 0, 1],
                     [0, 1, 0],
                     [1, 0, 0],
                     [0, 1, 1],
                     [1, 0, 1],
                     [1, 1, 0],
                     [1, 1, 1]])

faces = np.array([[0, 3, 4],
                  [7, 3, 4],
                  [0, 1, 4],
                  [0, 5, 7],
                  [3, 6, 5]
                  ])


def test_meshblock_instantiation():
    # check that meshblock can be instantiated properly
    block = MeshBlock(vertices, faces)
    assert isinstance(block, MeshBlock)
    assert hasattr(block, 'vertices')
    assert hasattr(block, 'faces')


def test_triangles():
    block = MeshBlock(vertices, faces)
    triangles = block.triangles
    assert triangles.shape[0] == faces.shape[0]
    assert triangles.shape[2] == 3


def test_midpoints():
    block = MeshBlock(vertices, faces)
    midpoints = block.midpoints
    assert midpoints.shape == (faces.shape[0], 3)



import numpy as np

from .orientedpointblock import OrientedPointBlock
from .propertyblock import PropertyBlock


class ParticleBlock(OrientedPointBlock):
    def __init__(self, positions: np.ndarray, orientations: np.ndarray, properties: dict, **kwargs):
        # Initialise OrientedPointBlock
        super().__init__(positions, orientations, **kwargs)

        # Add PropertyBlock
        self.properties = PropertyBlock(properties)

    def __shape_repr__(self):
        return f'{self.positions.data.shape}'



import re
from collections import defaultdict
from itertools import zip_longest

from ..utils import _path, listify
from .star import read_star
from .mrc import read_mrc
from .em import read_em

from ...core import DataCrate
from ...utils import AttributedList


# a mapping of file extensions to readers, tuple map to tuples:
#   - multiple extensions values in the keys use the same readers
#   - multiple readers are called in order from highest to lowers in case previous ones fail
# TODO: put this directly in the readers to make it plug and play?
readers = {
    ('.star',): (read_star,),
    ('.mrc',): (read_mrc,),
    ('.em',): (read_em,),
}


def read_file(file_path, **kwargs):
    """
    read a single file with the appropriate parser and return a list of datablocks
    """
    for ext, funcs in readers.items():
        if file_path.suffix in ext:
            for func in funcs:
                try:
                    datablocks = listify(func(file_path, **kwargs))
                    return datablocks
                except ValueError:
                    # this will be raised by individual readers when the file can't be read.
                    # Keep trying until all options are exhausted
                    continue
    raise ValueError(f'could not read {file_path}')


def find_files(paths, filters=None, recursive=False):
    """
    take a path or iterable thereof and find all the contained readable files
    filters: a regex-like strings or iterable thereof used to match filenames
    """
    # sanitize input
    paths = listify(paths)
    paths = [_path(path) for path in paths]
    filters = listify(filters)

    # extract files
    files = []
    for path in paths:
        if not path.exists():
            raise FileNotFoundError(f'{path} does not exist')

        # find all the readable files, if any
        if path.is_file():
            files.append(path)
        elif path.is_dir():
            basedir = '.'
            if recursive:
                basedir = '**'
            known_formats = [ext for formats in readers for ext in formats]
            for ext in known_formats:
                files.extend(file for file in path.glob(f'{basedir}/*{ext}'))
        if not files:
            raise FileNotFoundError(f'{path} does not contain any files of a know type')

    # filter files if requested
    files = [file for file in files if all(re.search(regex, str(file)) for regex in filters)]

    return files


def read_to_datablocks(paths, filters=None, recursive=False, strict=False, **kwargs):
    """
    read generic path(s) into the appropriate datablocks
    strict: if set to False, ignore failures and read what possible
    """
    datablocks = []
    for file in find_files(paths, filters=filters, recursive=recursive):
        try:
            datablocks.extend(read_file(file, **kwargs))
        except ValueError:
            if strict:
                raise
    if not datablocks:
        raise ValueError(f'could not read any data from {paths}')
    return datablocks


def read(paths, mode=None, **kwargs):
    """
    read generic path(s) and construct datacrates accordingly
    mode:
        - lone: each datablock in a separate crate
        - zip_by_type: crates with one of each datablock type
        - bunch: all datablocks in a single crate
    """
    modes = ('lone', 'zip_by_type', 'bunch')

    if mode is not None and mode not in modes:
        raise ValueError(f'mode can only be one of {modes}')

    datablocks = read_to_datablocks(paths, **kwargs)
    datablocks_by_type = defaultdict(list)
    for db in datablocks:
        datablocks_by_type[type(db)].append(db)

    if mode is None:
        # check if there are multiple types and lengths are all the same
        if len(datablocks_by_type) > 1:
            count_per_type = [len(db_type) for db_type in datablocks_by_type.values()]
            if len(set(count_per_type)) == 1:
                mode = 'zip_by_type'
            else:
                mode = 'lone'
        else:
            mode = 'lone'

    if mode == 'lone':
        crates = [DataCrate(db) for db in datablocks]
    elif mode == 'bunch':
        crates = [DataCrate(datablocks)]
    elif mode == 'zip_by_type':
        crates = []
        for dbs in zip_longest(datablocks_by_type.values()):
            crates.append(DataCrate(dbs))
        # TODO: add rescaling?

    return AttributedList(crates)



import numpy as np


def align_vectors(a: np.ndarray, b: np.ndarray):
    """
    Find array of rotation matrices r such that r @ a = b when a and b are arrays of normalised vectors
    implementation designed to avoid trig calls
    based on https://iquilezles.org/www/articles/noacos/noacos.htm
    :param a: normalised vector(s) of length 3
    :param b: normalised vector(s) of length 3
    :return: rotation matrix
    """
    # setup
    a = a.reshape(-1, 3)
    b = b.reshape(-1, 3)
    n_vectors = max(a.shape[0], b.shape[0])

    # cross product to find axis about which rotation should occur
    axis = np.cross(a, b, axis=1)
    # dot product equals cosine of angle between normalised vectors
    cos_angle = np.einsum('ij, ij -> i', a, b)
    # k is a constant which appears as a factor in the rotation matrix
    k = 1 / (1 + cos_angle)

    # construct rotation matrix
    r = np.empty((n_vectors, 3, 3))
    r[:, 0, 0] = (axis[:, 0] * axis[:, 0] * k) + cos_angle
    r[:, 0, 1] = (axis[:, 1] * axis[:, 0] * k) - axis[:, 2]
    r[:, 0, 2] = (axis[:, 2] * axis[:, 0] * k) + axis[:, 1]
    r[:, 1, 0] = (axis[:, 0] * axis[:, 1] * k) + axis[:, 2]
    r[:, 1, 1] = (axis[:, 1] * axis[:, 1] * k) + cos_angle
    r[:, 1, 2] = (axis[:, 2] * axis[:, 1] * k) - axis[:, 0]
    r[:, 2, 0] = (axis[:, 0] * axis[:, 2] * k) - axis[:, 1]
    r[:, 2, 1] = (axis[:, 1] * axis[:, 2] * k) + axis[:, 0]
    r[:, 2, 2] = (axis[:, 2] * axis[:, 2] * k) + cos_angle

    return r



import pandas as pd
import starfile
import eulerangles

from ...utils import _path, star_types


def matrix2euler(rotation_matrices, convention):
    """
    convert euler matrices to angles
    """
    # If rotation matrices represent rotations of particle onto reference
    # invert them so that we return matrices which transform reference onto particle
    if convention['rotate_reference'] is False:
        rotation_matrices = rotation_matrices.transpose((0, 2, 1))

    euler_angles = eulerangles.matrix2euler(rotation_matrices,
                                            target_axes=convention['axes'],
                                            target_intrinsic=convention['intrinsic'],
                                            target_positive_ccw=convention['positive_ccw'])
    return euler_angles


def write_star(particleblock, file_path, star_type='relion', data_colums=[], overwrite=False):
    """
    write a particle block to disk as a .star file
    star_type: one of the supported starfile types
    data_colums: particleblock proerties to include as columns
    """
    coords = particleblock.positions.data
    orientation_matrices = particleblock.orientations.data
    eulers = matrix2euler(orientation_matrices, star_types[star_type]['angle_convention'])
    properties = [values for name, values in particleblock.properties.items() if name in data_colums]

    df = pd.DataFrame()
    column_names = star_types[star_type]['coords'] + star_types[star_type]['angles'] + data_colums
    x, y, z = coords.T
    ang1, ang2, ang3 = eulers.T
    for name, values in zip(column_names, [x, y, z, ang1, ang2, ang3] + properties):
        df[name] = values

    path = str(_path(file_path))
    if not path.endswith('.star'):
        path = f'{path}.star'
    starfile.write(df, path, overwrite=overwrite)



from collections.abc import Iterable


class AttributedList(list):
    """
    a list that accepts dot notation to return attributes of its elements, if all
    contained elements have that attribute and itself does not

    You can also set attributes and call methods of the nested elements.
    If setattr is called with a value of type AttributedList, the list will attempt to set the attributes
    element by element
    """
    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError as e:
            try:
                attrs = AttributedList()
                for item in self:
                    # this lets us go as deep as needed, if working with nested lists
                    if isinstance(item, (list, AttributedList)):
                        for sub_item in AttributedList(item):
                            try:
                                attrs.extend(sub_item.__getattribute__(name))
                            # when finding methods, this is raised
                            # TODO: why?
                            except TypeError:
                                attrs.append(sub_item.__getattribute__(name))
                    else:
                        attrs.append(item.__getattribute__(name))
                return attrs
            except AttributeError:
                raise e

    def __setattr__(self, name, value):
        # check first in the normal way
        if hasattr(list(self), name):
            list.__setattr__(self, name, value)

        # can only safely do this recursively if we know where to find the atttributes lower down
        elif hasattr(self, name):
            zip_values = None

            # if another Attributed list is fed, do it value by value if possible
            if isinstance(value, AttributedList):
                # len(self) is not enough because it might be nested
                if len(self.__getattribute__(name)) == len(value):
                    zip_values = list(reversed(value))  # reversed for popping

            for item in self:
                if zip_values is not None:
                    # if an attributed list was given, use it value by value
                    value = zip_values.pop()
                # this lets us go as deep as needed, if working with nested lists
                if isinstance(item, Iterable):
                    for sub_item in AttributedList(item):
                        sub_item.__setattr__(name, value)
                else:
                    item.__setattr__(name, value)
        else:
            raise AttributeError


    def __call__(self, *args, **kwargs):
        return AttributedList(item.__call__(*args, **kwargs) for item in self)



import numpy as np

from .particleblock import ParticleBlock
from .pointblock import PointBlock
from .orientationblock import OrientationBlock


class TransformBlock(ParticleBlock):
    """
    A TransformBlock contains a set of 3D transforms described relative
    to their own fixed coordinate system, made up of...
    * positions : (n, 3) array of vectors describing shifts
      to apply as (dx, dy, dz) relative to the origin (0, 0, 0)
    * orientations : (n, 3, 3) array of rotation matrices which
      rotate (by premultiplication) the unit vectors (i, j, k)
    """

    def apply_on(self, particles: ParticleBlock):
        """
        Applies the transforms encoded in this TransformBlock on a
        set of particles in a ParticleBlock

        For m transforms on n particles this results in m x n transformed particles

        Parameters
        ----------
        particles : ParticleBlock

        Returns
        -------
        transformed_particles : ParticleBlock
                                Shifted, rotated particles

        """
        # set up necessary data, be explicit to avoid confusion
        particle_positions = particles.positions.data.reshape(-1, 3, 1)  # (n, 3, 1) array
        particle_orientations = particles.orientations  # (n, 3, 3) array of rotm
        shift_vectors = self.positions.data.reshape(-1, 1, 3, 1)  # (m, 1, 3, 1) array
        transform_orientations = self.orientations.data.reshape(-1, 1, 3, 3)  # (m, 1, 3, 3) array of rotm

        # Orient the shift vectors according to particle orientation
        oriented_shifts = particle_orientations @ shift_vectors  # (m, n, 3, 1)

        # Apply the transformed shifts onto the current particle positions
        output_positions = particle_positions + oriented_shifts  # (m, n, 3, 1)

        # Calculate the new orientations by composing rotations
        #  (n, 3, 3) @ (m, 1, 3, 3) -> (m, n, 3, 3)
        output_orientations = particle_orientations @ transform_orientations

        # replace nan filled matrices with original particle orientations
        # this happens when vectors are already aligned (causes div0)
        output_orientations = np.where(np.isnan(output_orientations), particle_orientations,
                                       output_orientations)

        # Create and return ParticleBlock from output
        points = PointBlock(output_positions.reshape((-1, 3)))
        orientations = OrientationBlock(output_orientations.reshape(-1, 3, 3))

        return ParticleBlock(points, orientations, {})



"""
Tests for DataBlock objects
"""

import pytest

from peepingtom.core import SimpleBlock, MultiBlock, DataCrate, Model
from ...test_data.blocks import pointblock, lineblock, orientationblock


def test_simpleblock():
    # assert that SimpleBlock class cannot be instantiated directly
    with pytest.raises(NotImplementedError):
        SimpleBlock([])

    # assert that subclassing and implementing _data_setter works
    class SubBlock(SimpleBlock):
        def __init__(self, data, **kwargs):
            super().__init__(data, **kwargs)

        def _data_setter(self, data):
            return data

    subblock = SubBlock([])
    assert isinstance(subblock, (SimpleBlock, SubBlock))

    # assert that subclass has parent attribute
    assert hasattr(subblock, 'parent')

    # assert that subclassing and not implementing _data_setter fails on subclass instantiation
    class SubBlock(SimpleBlock):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    with pytest.raises(TypeError):
        subblock = SubBlock()


def test_multiblock():
    block = MultiBlock()
    assert isinstance(block, MultiBlock)


def test_datacrate():
    # assert that datacrate instantiates properly
    crate = DataCrate([pointblock, lineblock, orientationblock])
    assert isinstance(crate, DataCrate)


def test_model():
    # assert that Model class cannot be instantiated directly
    with pytest.raises(TypeError):
        Model()



import numpy as np

from .pointblock import PointBlock


class MeshBlock(PointBlock):
    """
    Data structure for holding n-dimensional mesh data
    """
    def __init__(self, vertices: np.ndarray, faces: np.ndarray, **kwargs):
        """

        Parameters
        ----------
        vertices : array of shape (v, m)
                   v vertices in m dimensions
        faces : array of shape (f, 3)
                array of indices into the vertices defining connectivity
        kwargs : keyword arguments passed to SimpleBlock
        """
        super().__init__(data=vertices, **kwargs)
        self.vertices = vertices
        self.faces = faces
        self.data = self.vertices

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, vertices: np.ndarray):
        self._vertices = np.asarray(vertices)

    @property
    def faces(self):
        return self._faces

    @faces.setter
    def faces(self, faces: np.ndarray):
        faces = np.asarray(faces)
        if faces.shape[1] != 3:
            raise ValueError('shape of faces must be (f, 3)')
        self._faces = faces

    @property
    def triangles(self):
        triangles = self.vertices[self.faces]
        return triangles

    @property
    def midpoints(self):
        """
        Calculate the midpoints of each triangle in the mesh
        Returns
        -------
        midpoints : (n, m)
                    n midpoints in m dimensions
        """
        return self.triangles.mean(axis=1)



import numpy as np

from .base import Depictor


class ParticleDepictor(Depictor):
    def init_layers(self, point_kwargs={}, vector_kwargs={}):
        pkwargs = {'size': 3}
        vkwargs = {'length': 10}

        pkwargs.update(point_kwargs)
        vkwargs.update(vector_kwargs)

        p_layer = self.make_points_layer(self.datablock.positions.as_zyx(),
                                         name=f'{self.name} - particle positions',
                                         properties=self.datablock.properties.data,
                                         **pkwargs)
        self.layers.append(p_layer)

        # get positions and 'projection' vectors
        positions = self.datablock.positions.as_zyx()
        unit_z_rotated_order_xyz = self.datablock.orientations.oriented_vectors('z').reshape((-1, 3))
        unit_z_rotated_order_zyx = unit_z_rotated_order_xyz[:, ::-1]
        # attach appropriate higher dimensions indeces to vectors
        # TODO: make more general
        if self.datablock.positions.ndim > 3:
            padded = np.zeros_like(self.datablock.positions.data)
            padded[:, -3:] = unit_z_rotated_order_zyx
            unit_z_rotated_order_zyx = padded

        napari_vectors = np.stack([positions, unit_z_rotated_order_zyx], axis=1)
        v_layer = self.make_vectors_layer(napari_vectors,
                                          name=f'{self.name} - particle orientations',
                                          **vkwargs)
        self.layers.append(v_layer)

    @property
    def point_layer(self):
        return self.layers[0]

    @property
    def vector_layer(self):
        return self.layers[1]

    def update(self):
        try:
            self.point_layer.properties = {k: v for k, v in self.datablock.properties.data.items()
                                           if len(v) == len(self.datablock.positions.data)}
        except KeyError:
            # happens if layers do not exist: just pass
            pass



import numpy as np

from peepingtom.core.datablocks import LineBlock

# test data for lineblock
v = np.linspace(0, 12)
line_2d = np.column_stack([v, np.sin(v)])
line_3d = np.column_stack([v, np.sin(v), np.cos(v)])


def test_lineblock_instantiation():
    # test LineBlock instantiation
    LineBlock(line_2d)
    LineBlock(line_3d)


def test_lineblock_fit_spline():
    # test LineBlock.fit_spline
    block = LineBlock(line_3d)
    tck = block.fit_spline('xyz')

    assert block._tck is not None
    assert isinstance(tck, list)


def test_lineblock_evaluate_spline():
    # test LineBlock.evaluate_spline
    block = LineBlock(line_3d)
    block.fit_spline('xyz')
    for n in [10, 100, 1000]:
        spline = block.evaluate_spline(n)
        assert isinstance(spline, np.ndarray)

    assert block._tck is not None
    assert isinstance(block._tck, list)



import numpy as np

from .base import SimpleBlock


class SphereBlock(SimpleBlock):
    def __init__(self, center: np.ndarray, radius: float = None, **kwargs):
        super().__init__(**kwargs)
        # TODO: check if below works
        self.data = center, radius
        self.edge_point = None

    def _data_setter(self, center: np.ndarray, radius: float = None):
        self.center = center
        self.radius = float(radius)
        return self.center, self.radius

    @property
    def center(self):
        return self._center

    @center.setter
    def center(self, point: np.ndarray):
        return np.asarray(point).reshape(3)

    @property
    def edge_point(self):
        return self._edge_point

    @edge_point.setter
    def edge_point(self, edge_point: np.ndarray):
        edge_point = np.asarray(edge_point).reshape(3)
        self._edge_point = edge_point
        self._update_radius_from_edge_point()

    def _update_radius_from_edge_point(self):
        self.radius = np.linalg.norm(self.center - self.edge_point)



import numpy as np

from .base import MultiBlock
from .orientationblock import OrientationBlock
from .pointblock import PointBlock


class OrientedPointBlock(MultiBlock):
    def __init__(self, positions: np.ndarray, orientations: np.ndarray, **kwargs):
        super().__init__(**kwargs)
        self.positions = PointBlock(positions)
        self.orientations = OrientationBlock(orientations)

    def __shape_repr__(self):
        return f'{self.positions.data.shape}'



import mrcfile

from ...utils import _path


def write_mrc(imageblock, file_path, overwrite=False):
    """
    write an image block to disk as an .mrc file
    """
    path = str(_path(file_path))
    if not path.endswith('.mrc'):
        path = f'{path}.mrc'
    mrcfile.new(path, imageblock.data, overwrite=overwrite)



from abc import ABC, abstractmethod


class Alchemist(ABC):
    """
    Base class for all alchemist objects, used to transform a datablocks into other datablocks
    """
    def __init__(self, input_blocks):
        self.inputs = input_blocks
        self.outputs = {}

    def __repr__(self):
        return f'<{type(self).__name__}>'

    @abstractmethod
    def transform(self):
        """
        generate the output blocks based on the current settings and the input blocks
        """



from collections import Counter

from ...core import ImageBlock, ParticleBlock

from ..utils import _path, listify
from .mrc import write_mrc
from .em import write_em
from .star import write_star


writers = {
    (ImageBlock,): (write_mrc, write_em,),
    (ParticleBlock,): (write_star,),
}


def get_writer(datablock):
    for dbtypes, writer_funcs in writers.items():
        if type(datablock) in dbtypes:
            # TODO: implement some logic
            return writer_funcs[0]
    raise ValueError(f'could not find a writer for {type(datablock)}')


def write(datablocks, paths, overwrite=False, strict=False, **kwargs):
    """
    wite a list of datablocks to disk
    """
    # sanitize input
    datablocks = listify(datablocks)
    paths = [_path(path) for path in listify(paths)]

    # check if numbers match
    if len(datablocks) != len(paths) and len(datablocks):
        # were we given a dir path?
        if len(paths) != 1:
            raise NotImplementedError
        elif any(path.is_file() for path in paths):
            raise ValueError(f'number of datablocks and paths must be equal; '
                             f'got {len(datablocks)} datablocks and {len(paths)} paths')
        else:
            # i guess it's a dir. convert it to paths
            # TODO: not great with a single datablock
            dir_path = paths[0]
            dir_path.mkdir(parents=True, exist_ok=True)
            paths = []
            count_same_name = Counter()
            for datablock in datablocks:
                count_same_name[datablock.name] += 1
                if idx := count_same_name[datablock.name] > 1:
                    path = dir_path / datablock.name + str(idx)
                else:
                    path = dir_path / datablock.name
                paths.append(path)


    for datablock, path in zip(datablocks, paths):
        if not overwrite and path.exists():
            raise ValueError('file {path} exists!')
        try:
            writer = get_writer(datablock)
            writer(datablock, path, overwrite=overwrite, **kwargs)
        except ValueError:
            if strict:
                raise



from .base import Depictor


class LineDepictor(Depictor):
    def init_layers(self, point_kwargs={}, backbone_kwargs={}):
        # default keyword arguments
        pkwargs = {'size': 3,
                   'face_color': 'cornflowerblue'}
        bkwargs = {'edge_color': 'orangered',
                   'edge_width': 1}

        # update keyword arguments from passed dictionaries
        pkwargs.update(point_kwargs)
        bkwargs.update(backbone_kwargs)

        # get points data in napari expected order
        points_data = self.datablock.as_zyx()

        # get smooth backbone data in napari expected order
        backbone_data = self.datablock.smooth_backbone[:, ::-1]

        # draw points layer in napari
        points = self.make_points_layer(points_data,
                                        name=f'{self.name} - points',
                                        **pkwargs)

        self.layers.append(points)

        # draw line as path in napari
        backbone = self.make_shapes_layer(backbone_data,
                                          shape_type='path',
                                          name=f'{self.name} - line',
                                          **bkwargs)

        self.layers.append(backbone)

    @property
    def points_layer(self):
        return self.layers[0]

    @property
    def backbone_layer(self):
        return self.layers[1]

    def update(self):
        backbone_data = self.datablock.smooth_backbone[:, ::-1]
        self.backbone_layer.selected_data = {0}
        self.backbone_layer.remove_selected()
        self.backbone_layer.add(backbone_data, shape_type='path')

    def push_changes(self, event):
        self.datablock.data = self.points_layer.data[:, ::-1]



import starfile
import eulerangles

from ....core import ParticleBlock
from ...utils import star_types, guess_name


def split_and_name(dataframe, split_by, basename, name_regex=None):
    """
    splits a dataframe from a starfile in separate
    """
    if split_by in dataframe.columns:
        groups = dataframe.groupby('rlnMicrographName')
        return [(guess_name(subname, name_regex), dataframe) for subname, dataframe in groups]
    else:
        return [(guess_name(basename, name_regex), dataframe)]


def euler2matrix(euler_angles, convention):
    """
    convert euler angles to matrices
    """
    rotation_matrices = eulerangles.euler2matrix(euler_angles,
                                                 axes=convention['axes'],
                                                 intrinsic=convention['intrinsic'],
                                                 positive_ccw=convention['positive_ccw'])

    # If rotation matrices represent rotations of particle onto reference
    # invert them so that we return matrices which transform reference onto particle
    if convention['rotate_reference'] is False:
        rotation_matrices = rotation_matrices.transpose((0, 2, 1))
    return rotation_matrices


def read_star(star_path, data_columns=[], name_regex=None, **kwargs):
    """
    read a starfile into one or multiple ParticleBlocks
    data_columns: a list of column names to include as properties
    name_regex: a regex pattern to use to guess the name of individual datasets
    """
    raw_dfs = starfile.read((star_path), always_dict=True)

    # metadata = {}
    dataframes = []
    # check if relion3.1 format
    if list(raw_dfs.keys()) == ['optics', 'particles']:
        # TODO: use optics metadata
        # metadata = raw_dfs['optics']
        dataframes.append(raw_dfs['particles'])
    else:
        # assume every key is a different dataset (older version)
        # TODO: make more specific
        dataframes.extend(list(raw_dfs.values()))

    # create particles
    datablocks = []
    for dataframe in dataframes:
        # try every type of starfile convention
        for star_type, params in star_types.items():
            # check if the necessary columns exist
            if all(colname in dataframe.columns for colname in params['coords']):
                # split and guess names
                named_dfs = split_and_name(dataframe, split_by=params['split_by'],
                                           basename=star_path, name_regex=name_regex)
                for name, df in named_dfs:
                    # extract the data and convert it to peepingtom format
                    raw_positions = df[params['coords']]
                    raw_shifts = df.get(params['shifts'], 0)
                    raw_eulers = df[params['angles']]
                    raw_properties = df[data_columns]

                    positions = (raw_positions + raw_shifts).to_numpy()
                    orientations = euler2matrix(raw_eulers, params['angle_convention'])
                    properties = {key: df[key].to_numpy() for key in raw_properties}

                    datablocks.append(ParticleBlock(positions, orientations, properties, name=name))
                # break of out loop if something was found
                break
        else:
            raise ValueError(f'dataframe with columns {list(dataframe.columns)} could not be read')

    return datablocks
